--[[
Uber credits must be given to nymbia for the metatables and frames, thanks!
]]
--Ace2 Registering
ClassTimer = AceLibrary('AceAddon-2.0'):new('AceConsole-2.0', 'AceEvent-2.0', 'AceDB-2.0')

local ClassTimer = ClassTimer
local self = ClassTimer

local dewdrop = AceLibrary('Dewdrop-2.0')
local L = AceLibrary('AceLocale-2.2'):new('ClassTimer')
local sm = AceLibrary('SharedMedia-1.0')
local bs = AceLibrary('Babble-Spell-2.2')
local BS = setmetatable({}, {__index = function(self, v)
			local value
			if bs:HasTranslation(v) then
				value = bs[v]
			else
				value = v
				ClassTimer:Print("No Babble-Spell translation for: '"..v.."' in your locale, so this ability wont work!")
			end
			self[v] = value
			return value
end})

local unpack = unpack
local GetTime = GetTime
local table_sort = table.sort
local math_ceil = math.ceil
local gsub = gsub
local UnitIsUnit = UnitIsUnit

local locked = { target = true, player = true, focus = true, pet = true }

local new, del
do
	local cache = setmetatable({}, {__mode='k'})
	function new()
		local t = next(cache)
		if t then
			cache[t] = nil
			return t
		else
			return {}
		end
	end
	function del(t)
		for k in pairs(t) do
			t[k] = nil
		end
		cache[t] = true
		return nil
	end
end
local OnUpdate
do
	local function timenum(num)
		if num <= 10 then
			return ('%.1fs'):format(num)
		elseif num <= 60 then
			return ('%ds'):format(num)
		else
			return ('%dm'):format(math_ceil(num / 60))
		end
	end
	function OnUpdate(frame)
		local currentTime = GetTime()
		local endTime = frame.endTime
		if currentTime > endTime then
			if not locked[frame.unit] then locked[frame.unit] = true end
			self:UpdateUnitBars(frame.unit)
		else
			local remaining = (currentTime - frame.startTime)
			frame:SetValue(endTime - remaining)
			frame.timetext:SetText(timenum(endTime - currentTime))
			local sp = -frame:GetWidth()*remaining/(endTime-frame.startTime)
			frame.spark:SetPoint('CENTER', frame, 'RIGHT', sp, 0)
		end
		
	end
end

local function MouseUp(bar, button)
	if button == 'RightButton' then 
		local msg = L['%s has %s left']:format(bar.text:GetText(), bar.timetext:GetText()) 
		if UnitInRaid('player') then
			SendChatMessage(msg, 'RAID')
		elseif UnitInParty('player') then
			SendChatMessage(msg, 'PARTY')
		end
	end
end

local framefactory = {
	__index = function(t,k)
		local bar = CreateFrame('StatusBar', nil, UIParent)
		t[k] = bar
		bar:SetFrameStrata('MEDIUM')
		bar:Hide()
		bar:EnableMouse(true)
		bar:SetScript('OnUpdate', OnUpdate)
		bar:SetBackdrop({bgFile = 'Interface\\Tooltips\\UI-Tooltip-Background', tile = true, tileSize = 16})
		bar:SetMovable(true)
		bar.text = bar:CreateFontString(nil, 'OVERLAY')
		bar.timetext = bar:CreateFontString(nil, 'OVERLAY')
		bar.icon = bar:CreateTexture(nil, 'DIALOG')
		bar:SetScript('OnMouseUp', MouseUp)
		
		local spark = bar:CreateTexture(nil, "OVERLAY")
		bar.spark = spark
		spark:SetTexture("Interface\\CastingBar\\UI-CastingBar-Spark")
		spark:SetWidth(16)
		spark:SetBlendMode("ADD")
		spark:Show()
		self:ApplySettings()
		return bar
	end
}

local _, enClass = UnitClass("player")
local bars = {
	target = setmetatable({}, framefactory),
	focus = setmetatable({}, framefactory),
	player = setmetatable({}, framefactory),
	pet = enClass == "HUNTER" and setmetatable({}, framefactory) or nil
}

ClassTimer:RegisterDB('ClassTimerDB')

ClassTimer.defaults = {
	Abilities = {},
	Units = {	
		['**'] = {
			enable = true,
			buffs = true,
			debuffs = true,
			growup = false,
			showIcons = false,
			icons = true,
			iconSide = L['Left'],
			scale = 1,

			spacing = 0,
		},
	},
	nametext = true,
	timetext = true,
	
	AllInOne = false,
	texture = 'Blizzard',
	width = 150,
	height = 16,
	font = 'Friz Quadrata TT',
	fontsize = 9,
	alpha = 1,
	scale = 1,
	bartext = '%s (%a) (%u)',
	
	buffcolor = {0,0.49, 1, 1},
	debuffcolor = {1.0,0.7, 0, 1},
	backgroundcolor = {0,0, 0, 1},
		
	textcolor = {1,1,1},
}

 
function ClassTimer:OnInitialize()
--Revision 
	self.revision = tonumber(string.match('$Rev: 37703 $', '%d+'))
	self.version = self.revision and self.version .. ' |cffff8888r'..self.revision..'|r'
	local timerargs = {}
	local table =  self:CreateTimers()
	table['Trinkets'] = self:Trinkets()
	table['Race'] = self:Races()
	for k, v in pairs(table) do
		for n, m in pairs (v) do
			ClassTimer.defaults.Abilities[BS[m]] = true
			v[n] = BS[m]
		end
		local k = L[k]
		timerargs[k] = {
			type = 'text',
			order = 4,
			name = k,
			desc = L['Which Buffs to show.'],
			get = function(key) return self.db.profile.Abilities[key] end,
			set = function(key, value) self.db.profile.Abilities[key] = value end,
			validate = v,
			multiToggle = true,
		}
		
	end
	self.Menu.args.Timers = { type = 'group', order = 1, name = L['Timers'], desc = L['Enable or disable timers'], args = timerargs, disabled = '~IsActive' }
	self:RegisterDefaults('profile', self.defaults)
	self:RegisterChatCommand('/ClassTimerConsole', self.Menu)
	dewdrop:Register('ClassTimer', 'children', function() dewdrop:FeedAceOptionsTable(self.Menu) end)
	self:RegisterChatCommand('/ClassTimer', '/ct', function() dewdrop:Open('ClassTimer') end )
end

function ClassTimer:OnEnable()
	self:RegisterBucketEvent('UNIT_AURA', 0.5)
	self:RegisterEvent('PLAYER_TARGET_CHANGED')
	self:RegisterEvent('PLAYER_FOCUS_CHANGED')
	self:RegisterEvent('PLAYER_PET_CHANGED')
	self:RegisterEvent('SharedMedia_SetGlobal', function(mtype, override)
		if mtype == 'statusbar' then
			for _, k in pairs(bars) do
				for _, v in pairs(k) do
					v:SetStatusBarTexture(sm:Fetch('statusbar', override))
				end
			end
		end
	end)
	ClassTimer:ApplySettings()
end

function ClassTimer:OnDisable()
	for _, k in pairs(bars) do
		for _, v in pairs(k) do
			v:Hide()
		end
	end
end

function ClassTimer:OnDisable()
	for _, k in pairs(bars) do
		for _, v in pairs(k) do
			v:Hide()
		end
	end
end

function ClassTimer:PLAYER_TARGET_CHANGED()
	self:UpdateUnitBars('target')
end

function ClassTimer:PLAYER_FOCUS_CHANGED()
	self:UpdateUnitBars('focus')
end

function ClassTimer:PLAYER_PET_CHANGED()
	self:UpdateUnitBars('pet')
end

function ClassTimer:UNIT_AURA(units)
	for unit in pairs(units) do
		if self.db.profile.AllInOne then
			self:UpdateUnitBars('player')
			return
		end
		self:UpdateUnitBars(unit)
	end
end

do
	local function sortup(a,b)
			return a.remaining > b.remaining
	end
	local function sort(a,b)
			return a.remaining < b.remaining
	end
	local function text(spell, apps, unit)
		local str = self.db.profile.bartext
		str = gsub(str, '%%s', spell)
		str = gsub(str, '%%u', unit)
		str = gsub(str, '%%n', GetUnitName(unit))
		if apps and apps > 1 then
			str = gsub(str, '%%a', apps)
		else
			str = gsub(str, '%%a', '')
		end
		str = gsub(str, '%s%s+', ' ') 
		str = gsub(str, '%p%p+', '')
		str = gsub(str, '%s+$', '')		
		return str
	end
	local tmp = {}
	local called = false -- prevent recursive calls when new bars are created.
	function ClassTimer:GetBuffs(unit)
		if self.db.profile.Units[unit].buffs then
			for i = 1, 32 do
				local name, _, texture, count, duration, remaining = UnitBuff(unit, i)
				if not name then
					break
				end
				if duration and duration > 0 and self.db.profile.Abilities[name] then
					local t = new()
					tmp[#tmp+1] = t
					t.name = name
					t.unit = unit
					t.texture = texture
					t.duration = duration
					t.remaining = remaining
					t.count = count
					t.isbuff = true
				end
			end
		end
		if self.db.profile.Units[unit].debuffs then
			for i = 1, 40 do
				local name, _, texture, count, _, duration, remaining = UnitDebuff(unit, i)
				if not name then
					break
				end
				if duration and duration > 0 and self.db.profile.Abilities[name] then
					local t = new()
					tmp[#tmp+1] = t
					t.name = name
					t.unit = unit
					t.texture = texture
					t.duration = duration
					t.remaining = remaining
					t.count = count
				end
			end
		end
	end
	function ClassTimer:UpdateUnitBars(unit)
		if not bars[unit] then return end
		local db = self.db.profile
		if not locked[unit] then return end
		if db.AllInOne and unit ~= 'player' then unit = 'player' end
		if called then
			return
		end
		called = true
		
		if db.Units[unit].enable then
			local currentTime = GetTime()
			for k in pairs(tmp) do
				tmp[k] = del(tmp[k])
			end
			self:GetBuffs(unit)
			if db.AllInOne then
				if not UnitIsUnit('target', 'player') then self:GetBuffs('target') end
				if not UnitIsUnit('pet', 'player') then self:GetBuffs('pet') end
				if not UnitIsUnit('focus', 'player') and not UnitIsUnit('focus', 'target') then self:GetBuffs('focus') end
			end
			table_sort(tmp, db.Units[unit].growup and sortup or sort)
			for k,v in ipairs(tmp) do
				local bar = bars[unit][k]
				
				bar.text:SetText(text(v.name, v.count, v.unit))
				bar.icon:SetTexture(v.texture)
				local elapsed = (v.duration - v.remaining)
				local startTime, endTime = (currentTime - elapsed), (currentTime + v.remaining)
				bar.startTime = startTime
				bar.unit = unit
				bar.endTime = endTime
				bar.isbuff = v.isbuff
				if not db.Units[unit].showIcons then
					if bar.isbuff then
						bar:SetStatusBarColor(unpack(db.buffcolor))
					else
						bar:SetStatusBarColor(unpack(db.debuffcolor))
					end
				end
				bar:SetMinMaxValues(startTime, endTime)
				bar:Show()
			end
			for i = #tmp+1, #bars[unit] do
				bars[unit][i]:Hide()
			end
		else
			for _, v in ipairs(bars[unit]) do
				v:Hide()
			end
		end
		called = false
	end
end
do
	local function apply(unit, i, bar, db)
		local position, icons, ShowIcons, iconside, gap, spacing, direction, width
		local bars = bars[unit]
		growup = db.Units[unit].growup
		icons = db.Units[unit].icons
		showIcons = db.Units[unit].showIcons 
		iconSide = db.Units[unit].iconSide
		gap = db.Units[unit].gap
		spacing = db.Units[unit].spacing
		offset = db.Units[unit].offset
		bar:ClearAllPoints()
		bar:SetStatusBarTexture(sm:Fetch('statusbar', db.texture))
		bar:SetHeight(db.height)
		bar:SetScale(db.Units[unit].scale)
		bar:SetAlpha(db.alpha)
		bar:SetWidth(db.width)
		bar.spark:SetHeight(db.height + 25)
		if showIcons then
			bar:SetStatusBarColor(1, 1, 1, 0)
			bar:SetBackdropColor(1, 1, 1, 0)
			bar.spark:Hide()
			bar.text:SetFont(sm:Fetch('font', db.font), db.fontsize)
		else 
			if bar.isbuff then
				bar:SetStatusBarColor(unpack(db.buffcolor))
			else
				bar:SetStatusBarColor(unpack(db.debuffcolor))
			end
			bar.spark:Show()
			bar:SetBackdropColor(unpack(db.backgroundcolor))
		end
		if i == 1 then
			if db.Units[unit].x then
				bar:SetPoint('BOTTOMLEFT', UIParent, 'BOTTOMLEFT', db.Units[unit].x, db.Units[unit].y)
			else
				bar:SetPoint('CENTER', UIParent)
			end
		else
			if growup then
				bar:SetPoint('BOTTOM', bars[i-1], 'TOP', 0, spacing)
			else 
				bar:SetPoint('TOP', bars[i-1], 'BOTTOM', 0, -1 * spacing)
			end
		end
		
		local timetext = bar.timetext
		if db.timetext then
			timetext:Show()
			timetext:ClearAllPoints()
			timetext:SetFont(sm:Fetch('font', db.font), db.fontsize)
			timetext:SetShadowColor( 0, 0, 0, 1)
			timetext:SetShadowOffset( 0.8, -0.8 )
			timetext:SetTextColor(unpack(db.textcolor))
			timetext:SetNonSpaceWrap(false)
			timetext:ClearAllPoints()
			timetext:SetWidth(db.width)
			timetext:SetPoint('LEFT', bar, 'LEFT', showIcons and 2 or -2, 0)
			timetext:SetJustifyH(showIcons and 'LEFT' or 'RIGHT')
		else
			timetext:Hide()
		end
		
		local text = bar.text
		text:SetFont(sm:Fetch('font', db.font), db.fontsize)
		if db.nametext and not showIcons then
			text:Show()
			text:ClearAllPoints()
			text:SetShadowColor( 0, 0, 0, 1)
			text:SetShadowOffset( 0.8, -0.8 )
			text:SetTextColor(unpack(db.textcolor))
			text:SetNonSpaceWrap(false)
			text:ClearAllPoints()
			text:SetWidth(db.width - db.fontsize * 2.5)
			text:SetHeight(db.height)
			text:SetPoint('LEFT', bar, 'LEFT', 2, 0)
			text:SetJustifyH('LEFT')
		else
			text:Hide()
		end

		local icon = bar.icon
		if icons then
			icon:Show()
			icon:SetWidth(db.height)
			icon:SetHeight(db.height)
			icon:SetTexCoord(0.07, 0.93, 0.07, 0.93)
			icon:ClearAllPoints()
			if iconSide == L['Left'] then
				icon:SetPoint('RIGHT', bar, 'LEFT', 0, 0)
			else
				icon:SetPoint('LEFT', bar, 'RIGHT', 0, 0)
			end
		else
			icon:Hide()
		end
		
		return direction
	end
	function ClassTimer:ApplySettings()
		for n, k in pairs(bars) do
			for u, v in pairs(k) do
				apply(n, u, v, ClassTimer.db.profile)
			end
		end
	end
end 

do
	local function dragstart()
		this:StartMoving()
	end
	local function dragstop()
		ClassTimer.db.profile.Units[this.unit].x = this:GetLeft()
		ClassTimer.db.profile.Units[this.unit].y = this:GetBottom()
		this:StopMovingOrSizing()
	end
	local function set(table, value)
		local s1, s2 = unpack(table)
		if s2 then
			ClassTimer.db.profile.Units[s1][s2] = value
		else
			ClassTimer.db.profile[s1] = value
		end
			ClassTimer.ApplySettings()
	end
	local function get(table)
		local s1, s2 = unpack(table)
		if s2 then
			return ClassTimer.db.profile.Units[s1][s2]
		else
			return ClassTimer.db.profile[s1]
		end
	end
	local function setcolor(field, ...)
		ClassTimer.db.profile[field] = {...}
		ClassTimer.ApplySettings()
	end
	local function getcolor(field)
		return unpack(ClassTimer.db.profile[field])
	end 
	local function hideoptions(unit)	
		return not ClassTimer.db.profile.Units[unit[1]].enable
	end
	local function Unit(unit)
		return bars[unit] and {
			type = 'group',
			name = L[unit],
			desc = L[unit],
			order = 1,
			disabled = function() return unit ~= 'player' and ClassTimer.db.profile.AllInOne end,
			args = {
				show = {
					type = 'toggle',
					name = L['Enable %s']:format(L[unit]),
					desc = L['Show buffs/debuffs for your %s']:format(L[unit]),
					get = get,
					set = set,
					passValue = {unit, 'enable'},
					order = 1,
				},
				space = {
					type = 'header',
					order = 2,
					passValue = {unit},
					hidden = hideoptions,
				},
				buffs = {
					type = 'toggle',
					name = L['Enable Buffs'],
					desc = L['Show buffs for your %s']:format(L[unit]),
					get = get,
					set = set,
					order = 3,
					passValue = {unit, 'buffs'},
					hidden = hideoptions,
				},
				debuffs = {
					type = 'toggle',
					name = L['Enable Debuffs'],
					desc = L['Show debuffs for your %s']:format(L[unit]),
					get = get,
					order = 4,
					set = set,
					passValue = {unit, 'debuffs'},
					hidden = hideoptions,
				},
				lock = {
					type = 'toggle',
					name = L['Lock'],
					desc = L['Toggle Cast Bar lock'],
					order = 5,
					hidden = hideoptions,
					passValue = {unit},
					get = function()
						return locked[unit]
					end,
					set = function(_, v)
						if v then
							bars[unit][1]:SetScript('OnDragStart', nil)
							bars[unit][1]:SetScript('OnDragStop', nil)
						else
							local bar = bars[unit][1]
							bar.icon:SetTexture("Interface\\Icons\\Ability_Druid_Enrage")
							bar.text:SetText(L['Drag to move'])
							bar.startTime = GetTime()
							bar.endTime = GetTime() + 120
							bar:SetMinMaxValues(GetTime(),  GetTime() + 120)
							bar:RegisterForDrag('LeftButton')
							bar.unit = unit
							bar:SetScript('OnDragStart', dragstart)
							bar:SetScript('OnDragStop', dragstop)
							bar:SetAlpha(1)
							bar:Show()
						end
						locked[unit] = v
						ClassTimer:UpdateUnitBars(unit)
					end,
				},
				scale = {
					type = 'range',
					name = L['Scale'],
					desc = L['Set the scale of the bars for your %s']:format(L[unit]),
					get = get,
					set = set,
					min = .1,
					max = 2,
					step = .01,
					bigStep = .1,
					order = 12,
					passValue = {unit, 'scale'},
					hidden = hideoptions,
				},
				spacing = {
					type = 'range',
					name = L['Spacing'],
					desc = L['Tweak the space between bars for your %s']:format(L[unit]),
					min = -5,
					max = 5,
					step = .1,
					bigStep = 1,
					get = get,
					set = set,
					passValue = {unit, 'spacing'},
					order = 13,
					hidden = hideoptions,
				},
				iconside = {
					type = 'text',
					name = L['Icon Position'],
					desc = L['Set the side of the bar that the icon appears on'],
					get = get,
					set = set,
					validate = {L['Left'], L['Right']},
					passValue = {unit, 'iconSide'},
					order = 20,
					hidden = hideoptions,
				},
				showicons = {
					type = 'toggle',
					name = L['Show Icons'],
					desc = L['Show icons on buffs and debuffs for your %s']:format(L[unit]),
					get = get,
					set = set,
					passValue = {unit, 'icons'},
					order = 30,
					hidden = hideoptions,
				},
				showiconsOnly = {
					type = 'toggle',
					name = L['Show Only Icons'],
					desc = L['Show only icons and timeleft for your %s']:format(L[unit]),
					get = get,
					set = set,
					passValue = {unit, 'showIcons'},
					order = 31,
					hidden = hideoptions,
				},
				growUp = {
					type = 'toggle',
					name = L['Grow Up'],
					desc = L['Set %s bars to grow up']:format(L[unit]),
					get = get,
					set = set,
					order = 32,
					passValue = {unit, 'growup'},
					hidden = hideoptions,
				},
			}
		} or nil
	end

	ClassTimer.Menu = { 
	    type ='group',
	    args = {
			spacer = {
				type = 'header',
				order = 3,
			},
			BarSettings = {
				type = 'group',
				name = L['Bar Settings'],
				desc = L['Bar Settings'],
				order = 3,
				disabled = '~IsActive',
				args = {
					focus = Unit('focus'),
					target = Unit('target'),
					player = Unit('player'),
					pet = Unit('pet'),
					spacer = {
						type = 'header',
						order = 19,
					},
					alpha = {
						type = 'range',
						name = L['Alpha'],
						desc = L['Set the alpha of the bars'],
						min = 0.05,
						max = 1,
						step = 0.01,
						bigStep = .1,
						isPercent = true,
						get = get,
						set = set,
						passValue = {'alpha'},
						order = 20,
					},
					fontsize = {
						type = 'range',
						name = L['Font Size'],
						desc = L['Set the font size for the bars'],
						min = 3,
						max = 15,
						step = 2,
						bigStep = .05,
						get = get,
						set = set,
						passValue = {'fontsize'},
						order = 21,
					},
					height = {
						type = 'range',
						name = L['Bar Height'],
						desc = L['Set the height of the bars'],
						min = 4,
						max = 25,
						step = 2,
						bigStep = 5,
						get = get,
						set = set,
						passValue = {'height'},
						order = 22,
					},
					width = {
						type = 'range',
						name = L['Bar Width'],
						desc = L['Set the width of the bars'],
						min = 50,
						max = 300,
						step = 2,
						bigStep = 5,
						get = get,
						set = set,
						passValue = {'width'},
						order = 23,
					},
					font = {
						type = 'text',
						name = L['Font'],
						desc = L['Set the font used in the bars'],
						validate = sm:List('font'),
						get = get,
						set = set,
						passValue = {'font'},
						order = 40,
					},
					texture = {
						type = 'text',
						name = L['Texture'],
						desc = L['Set the bar Texture'],
						validate = sm:List('statusbar'),
						order = 41,
						get = get,
						set = set,
						passValue = {'texture'},
					},
					text = {
						type = 'text',
						name = L['Bar Text'],
						desc = L['Set the bar text'],
						get = get,
						set = set,
						order = 42,
						usage = L['<%s for spell, %a for applications, %n for name, %u for unit. %s must be included>'],
						validate = function(name) return string.find(name, '%%s') end,
						passValue = {'bartext'},
					},
					textcolor = {
						type = 'color',
						name = L['Text Color'],
						desc = L['Set the color of the text for the bars'],
						get = getcolor,
						set = setcolor,
						passValue = 'textcolor',
						order = 60,
					},
					buffcolor = {
						type = 'color',
						name = L['Buff Color'],
						desc = L['Set the color of the bars for buffs'],
						get = getcolor,
						set = setcolor,
						hasAlpha = true,
						passValue = 'buffcolor',
						order = 61,
					},
					debuffcolor = {
						type = 'color',
						name = L['Debuff Color'],
						desc = L['Set the color of the bars for debuffs'],
						get = getcolor,
						set = setcolor,
						hasAlpha = true,
						passValue = 'debuffcolor',
						order = 62,
					},
					backgroundcolor = {
						type = 'color',
						name = L['Background Color'],
						desc = L['Set the color of the bars background'],
						get = getcolor,
						set = setcolor,
						hasAlpha = true,
						passValue = 'backgroundcolor',
						order = 62,
					},
					AllInOne = {
						type = 'toggle',
						name = L['AllInOne'],
						desc = L['Display all the buffs and debuffs on the "player" bar'],
						get = get,
						set = set,
						passValue = {'AllInOne'},
						order = 80,
					},
					nametext = {
						type = 'toggle',
						name = L['Name Text'],
						desc = L['Display the names of buffs/debuffs on their bars'],
						get = get,
						set = set,
						passValue = {'nametext'},
						order = 80,
					},
					timetext = {
						type = 'toggle',
						name = L['Time Text'],
						desc = L['Display the time remaining on buffs/debuffs on their bars'],
						get = get,
						set = set,
						passValue = {'timetext'},
						order = 81,
					},
				},
			}
		}
	}
end
